@@ -403,7 +403,12 @@ void RebuildMap(void)
 			close(dfd);
 			if (ok != 0)
 				continue;
-			info = st->ss->container_content(st, subarray);
+			if (subarray)
+				info = st->ss->container_content(st, subarray);
+			else {
+				info = malloc(sizeof(*info));
+				st->ss->getinfo_super(st, info, NULL);
+			}
 			if (!info)
 				continue;
 
@@ -417,7 +422,7 @@ void RebuildMap(void)
 				 * an MD_DEVNAME for udev.
 				 * The name needs to be unique both in /dev/md/
 				 * and in this mapfile.
-				 * It needs to match watch -I or -As would come
+				 * It needs to match what -I or -As would come
 				 * up with.
 				 * That means:
 				 *   Check if array is in mdadm.conf 
