@@ -210,7 +210,10 @@ int create_mddev(char *dev, char *name, int autof, int trustworthy,
 			char *ep;
 			if (cname[0] == 'd')
 				sp++;
-			num = strtoul(sp, &ep, 10);
+			if (isdigit(sp[0]))
+				num = strtoul(sp, &ep, 10);
+			else
+				ep = sp;
 			if (ep == sp || *ep || num < 0)
 				num = -1;
 			else if (cname[0] == 'd')
@@ -289,8 +292,17 @@ int create_mddev(char *dev, char *name, int autof, int trustworthy,
 		int cnlen;
 		strncpy(cname, name, 200);
 		cname[200] = 0;
-		while ((cp = strchr(cname, '/')) != NULL)
-			*cp = '-';
+		for (cp = cname; *cp ; cp++)
+			switch (*cp) {
+			case '/':
+				*cp = '-';
+				break;
+			case ' ':
+			case '\t':
+				*cp = '_';
+				break;
+			}
+
 		if (trustworthy == LOCAL ||
 		    (trustworthy == FOREIGN && strchr(cname, ':') != NULL)) {
 			/* Only need suffix if there is a conflict */
