@@ -445,7 +445,7 @@ static int update_super0(struct supertype *st, struct mdinfo *info,
 		       sb32+MD_SB_GENERIC_CONSTANT_WORDS+7+1,
 		       (MD_SB_WORDS - (MD_SB_GENERIC_CONSTANT_WORDS+7+1))*4);
 		if (verbose >= 0)
-			fprintf (stderr, Name ": adjusting superblock of %s for 2.2/sparc compatability.\n",
+			fprintf (stderr, Name ": adjusting superblock of %s for 2.2/sparc compatibility.\n",
 				 devname);
 	} else if (strcmp(update, "super-minor") ==0) {
 		sb->md_minor = info->array.md_minor;
@@ -633,9 +633,9 @@ static int init_super0(struct supertype *st, mdu_array_info_t *info,
 	sb->gvalid_words = 0; /* ignored */
 	sb->ctime = time(0);
 	sb->level = info->level;
-	if (size != (unsigned long long)info->size)
+	sb->size = size;
+	if (size != (unsigned long long)sb->size)
 		return 0;
-	sb->size = info->size;
 	sb->nr_disks = info->nr_disks;
 	sb->raid_disks = info->raid_disks;
 	sb->md_minor = info->md_minor;
@@ -657,7 +657,7 @@ static int init_super0(struct supertype *st, mdu_array_info_t *info,
 		if (rfd >= 0)
 			close(rfd);
 	}
-	if (homehost) {
+	if (homehost && !uuid) {
 		char buf[20];
 		char *hash = sha1_buffer(homehost,
 					 strlen(homehost),
