@@ -32,6 +32,8 @@
 #include	<dirent.h>
 #include	<signal.h>
 
+int __offroot;
+
 /*
  * following taken from linux/blkpg.h because they aren't
  * anywhere else and it isn't safe to #include linux/ * stuff.
@@ -146,16 +148,16 @@ int get_linux_version()
 {
 	struct utsname name;
 	char *cp;
-	int a,b,c;
+	int a = 0, b = 0,c = 0;
 	if (uname(&name) <0)
 		return -1;
 
 	cp = name.release;
 	a = strtoul(cp, &cp, 10);
-	if (*cp != '.') return -1;
-	b = strtoul(cp+1, &cp, 10);
-	if (*cp != '.') return -1;
-	c = strtoul(cp+1, NULL, 10);
+	if (*cp == '.')
+		b = strtoul(cp+1, &cp, 10);
+	if (*cp == '.')
+		c = strtoul(cp+1, &cp, 10);
 
 	return (a*1000000)+(b*1000)+c;
 }
@@ -192,6 +194,7 @@ long long parse_size(char *size)
 	 * followed by 'K', 'M', or 'G'.
 	 * Without a suffix, K is assumed.
 	 * Number returned is in sectors (half-K)
+	 * -1 returned on error.
 	 */
 	char *c;
 	long long s = strtoll(size, &c, 10);
@@ -213,7 +216,7 @@ long long parse_size(char *size)
 		}
 	}
 	if (*c)
-		s = 0;
+		s = -1;
 	return s;
 }
 
@@ -309,10 +312,15 @@ int test_partition_from_id(dev_t id)
 	return rv;
 }
 
-int enough(int level, int raid_disks, int layout, int clean,
-	   char *avail, int avail_disks)
+int enough(int level, int raid_disks, int layout, int clean, char *avail)
 {
 	int copies, first;
+	int i;
+	int avail_disks = 0;
+
+	for (i = 0; i < raid_disks; i++)
+		avail_disks += !!avail[i];
+
 	switch (level) {
 	case 10:
 		/* This is the tricky one - we need to check
@@ -324,14 +332,15 @@ int enough(int level, int raid_disks, int layout, int clean,
 			/* there must be one of the 'copies' form 'first' */
 			int n = copies;
 			int cnt=0;
+			int this = first;
 			while (n--) {
-				if (avail[first])
+				if (avail[this])
 					cnt++;
-				first = (first+1) % raid_disks;
+				this = (this+1) % raid_disks;
 			}
 			if (cnt == 0)
 				return 0;
-
+			first = (first+(layout&255)) % raid_disks;
 		} while (first != 0);
 		return 1;
 
@@ -363,14 +372,14 @@ int enough_fd(int fd)
 	struct mdu_array_info_s array;
 	struct mdu_disk_info_s disk;
 	int avail_disks = 0;
-	int i;
+	int i, rv;
 	char *avail;
 
 	if (ioctl(fd, GET_ARRAY_INFO, &array) != 0 ||
 	    array.raid_disks <= 0)
 		return 0;
 	avail = calloc(array.raid_disks, 1);
-	for (i=0; i < 1024 && array.nr_disks > 0; i++) {
+	for (i=0; i < MAX_DISKS && array.nr_disks > 0; i++) {
 		disk.number = i;
 		if (ioctl(fd, GET_DISK_INFO, &disk) != 0)
 			continue;
@@ -386,9 +395,10 @@ int enough_fd(int fd)
 		avail[disk.raid_disk] = 1;
 	}
 	/* This is used on an active array, so assume it is clean */
-	return enough(array.level, array.raid_disks, array.layout,
-		      1,
-		      avail, avail_disks);
+	rv = enough(array.level, array.raid_disks, array.layout,
+		    1, avail);
+	free(avail);
+	return rv;
 }
 
 
@@ -535,6 +545,7 @@ int check_raid(int fd, char *name)
 	struct supertype *st = guess_super(fd);
 
 	if (!st) return 0;
+	st->ignore_hw_compat = 1;
 	st->ss->load_super(st, fd, name);
 	/* Looks like a raid array .. */
 	fprintf(stderr, Name ": %s appears to be part of a raid array:\n",
@@ -639,7 +650,7 @@ char *human_size(long long bytes)
 	 * We allow upto 2048Megabytes before converting to
 	 * gigabytes, as that shows more precision and isn't
 	 * too large a number.
-	 * Terrabytes are not yet handled.
+	 * Terabytes are not yet handled.
 	 */
 
 	if (bytes < 5000*1024)
@@ -701,6 +712,14 @@ void print_r10_layout(int layout)
 
 unsigned long long calc_array_size(int level, int raid_disks, int layout,
 				   int chunksize, unsigned long long devsize)
+{
+	if (level == 1)
+		return devsize;
+	devsize &= ~(unsigned long long)((chunksize>>9)-1);
+	return get_data_disks(level, layout, raid_disks) * devsize;
+}
+
+int get_data_disks(int level, int layout, int raid_disks)
 {
 	int data_disks = 0;
 	switch (level) {
@@ -712,8 +731,8 @@ unsigned long long calc_array_size(int level, int raid_disks, int layout,
 	case 10: data_disks = raid_disks / (layout & 255) / ((layout>>8)&255);
 		break;
 	}
-	devsize &= ~(unsigned long long)((chunksize>>9)-1);
-	return data_disks * devsize;
+
+	return data_disks;
 }
 
 #if !defined(MDASSEMBLE) || defined(MDASSEMBLE) && defined(MDASSEMBLE_AUTO)
@@ -778,10 +797,14 @@ int find_free_devnum(int use_partitions)
 	     devnum = devnum ? devnum-1 : (1<<20)-1) {
 		char *dn;
 		int _devnum;
+		char nbuf[50];
 
 		_devnum = use_partitions ? (-1-devnum) : devnum;
 		if (mddev_busy(_devnum))
 			continue;
+		sprintf(nbuf, "%s%d", use_partitions?"mdp":"md", devnum);
+		if (!conf_name_is_free(nbuf))
+			continue;
 		/* make sure it is new to /dev too, at least as a
 		 * non-standard */
 		dn = map_dev(dev2major(_devnum), dev2minor(_devnum), 0);
@@ -951,9 +974,10 @@ struct supertype *super_by_fd(int fd, char **subarrayp)
 		char *dev = verstr+1;
 
 		subarray = strchr(dev, '/');
-		if (subarray)
+		if (subarray) {
 			*subarray++ = '\0';
-		subarray = strdup(subarray);
+			subarray = strdup(subarray);
+		}
 		container = devname2devnum(dev);
 		if (sra)
 			sysfs_free(sra);
@@ -1120,7 +1144,8 @@ static int get_gpt_last_partition_end(int fd, unsigned long long *endofpart)
 {
 	struct GPT gpt;
 	unsigned char empty_gpt_entry[16]= {0};
-	struct GPT_part_entry part;
+	struct GPT_part_entry *part;
+	char buf[512];
 	unsigned long long curr_part_end;
 	unsigned all_partitions, entry_size;
 	unsigned part_nr;
@@ -1144,18 +1169,20 @@ static int get_gpt_last_partition_end(int fd, unsigned long long *endofpart)
 
 	/* sanity checks */
 	if (all_partitions > 1024 ||
-	    entry_size > 512)
+	    entry_size > sizeof(buf))
 		return -1;
 
+	part = (struct GPT_part_entry *)buf;
+
 	for (part_nr=0; part_nr < all_partitions; part_nr++) {
 		/* read partition entry */
-		if (read(fd, &part, entry_size) != (ssize_t)entry_size)
+		if (read(fd, buf, entry_size) != (ssize_t)entry_size)
 			return 0;
 
 		/* is this valid partition? */
-		if (memcmp(part.type_guid, empty_gpt_entry, 16) != 0) {
+		if (memcmp(part->type_guid, empty_gpt_entry, 16) != 0) {
 			/* check the last lba for the current partition */
-			curr_part_end = __le64_to_cpu(part.ending_lba);
+			curr_part_end = __le64_to_cpu(part->ending_lba);
 			if (curr_part_end > *endofpart)
 				*endofpart = curr_part_end;
 		}
@@ -1253,7 +1280,7 @@ void get_one_disk(int mdfd, mdu_array_info_t *ainf, mdu_disk_info_t *disk)
 	int d;
 
 	ioctl(mdfd, GET_ARRAY_INFO, ainf);
-	for (d = 0 ; d < 1024 ; d++) {
+	for (d = 0 ; d < MAX_DISKS ; d++) {
 		if (ioctl(mdfd, GET_DISK_INFO, disk) == 0 &&
 		    (disk->major || disk->minor))
 			return;
@@ -1288,6 +1315,20 @@ int open_container(int fd)
 			continue;
 		if (de->d_name[0] == '.')
 			continue;
+		/* Need to make sure it is a container and not a volume */
+		sprintf(e, "/%s/md/metadata_version", de->d_name);
+		dfd = open(path, O_RDONLY);
+		if (dfd < 0)
+			continue;
+		n = read(dfd, buf, sizeof(buf));
+		close(dfd);
+		if (n <= 0 || (unsigned)n >= sizeof(buf))
+			continue;
+		buf[n] = 0;
+		if (strncmp(buf, "external", 8) != 0 ||
+		    n < 10 ||
+		    buf[9] == '/')
+			continue;
 		sprintf(e, "/%s/dev", de->d_name);
 		dfd = open(path, O_RDONLY);
 		if (dfd < 0)
@@ -1369,7 +1410,7 @@ int open_subarray(char *dev, char *subarray, struct supertype *st, int quiet)
 		if (!quiet)
 			fprintf(stderr, Name ": Couldn't open %s, aborting\n",
 				dev);
-		return 2;
+		return -1;
 	}
 
 	st->devnum = fd2devnum(fd);
@@ -1572,7 +1613,7 @@ int mdmon_running(int devnum)
 
 int start_mdmon(int devnum)
 {
-	int i;
+	int i, skipped;
 	int len;
 	pid_t pid;	
 	int status;
@@ -1587,7 +1628,7 @@ int start_mdmon(int devnum)
 	if (check_env("MDADM_NO_MDMON"))
 		return 0;
 
-	len = readlink("/proc/self/exe", pathbuf, sizeof(pathbuf));
+	len = readlink("/proc/self/exe", pathbuf, sizeof(pathbuf)-1);
 	if (len > 0) {
 		char *sl;
 		pathbuf[len] = 0;
@@ -1603,13 +1644,25 @@ int start_mdmon(int devnum)
 	switch(fork()) {
 	case 0:
 		/* FIXME yuk. CLOSE_EXEC?? */
-		for (i=3; i < 100; i++)
-			close(i);
+		skipped = 0;
+		for (i=3; skipped < 20; i++)
+			if (close(i) < 0)
+				skipped++;
+			else
+				skipped = 0;
+
 		for (i=0; paths[i]; i++)
-			if (paths[i][0])
-				execl(paths[i], "mdmon",
-				      devnum2devname(devnum),
-				      NULL);
+			if (paths[i][0]) {
+				if (__offroot) {
+					execl(paths[i], "mdmon", "--offroot",
+					      devnum2devname(devnum),
+					      NULL);
+				} else {
+					execl(paths[i], "mdmon",
+					      devnum2devname(devnum),
+					      NULL);
+				}
+			}
 		exit(1);
 	case -1: fprintf(stderr, Name ": cannot run mdmon. "
 			 "Array remains readonly\n");
@@ -1697,7 +1750,8 @@ int experimental(void)
 	if (check_env("MDADM_EXPERIMENTAL"))
 		return 1;
 	else {
-		fprintf(stderr, Name ": To use this feature MDADM_EXPERIMENTAL enviroment variable has to defined.\n");
+		fprintf(stderr, Name ": To use this feature MDADM_EXPERIMENTAL"
+				" environment variable has to be defined.\n");
 		return 0;
 	}
 }
